/*
===============================================================================

  FILE:  hexahedronrotations.h

  CONTENTS:

    utility function to rotate a hexahedron.

    The used numbering is the CUBIT numbering (http://cubit.sandia.gov/help-version11.0/cubithelp.htm)
    	7-------6
       /.      /|
      4-------5 |
      | .     | |
      | 3.....|.2
      |.      |/
      0-------1

    The faces are numbered as following:
	- face 0 at the bottom (0,1,2,3)
	- face 1 on the left (0,3,7,4)
	- face 2 in front (0,4,5,1)
	- face 3 on the right (1,5,6,2)
	- face 4 in the back (2,6,7,3)
	- face 5 at the top (4,7,6,5)

    The halfedges representing each face are:
	- 1->0 for face 0
	- 3->0 for face 1
	- 4->0 for face 2
	- 5->1 for face 3
	- 6->2 for face 4
	- 7->4 for face 5

    All the following functions rotate a hehahedron about one of the X, Y, or Z axis

	    Z   
	    |  X
	    | /
	    |/_______Y


    RX conserves faces 2 and 4
    RY conserves faces 1 and 3
    RZ conserves faces 0 and 5

  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009  clement courbet

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    06 October 2009 -- initial version

===============================================================================
*/

#ifndef HEXAHEDRON_ROTATIONS_H
#define HEXAHEDRON_ROTATIONS_H

/**
 * rotate the hexahedron a quarter (Q) of a turn about the Z axis in the trigo direction
 * @param currentRotation the current rotation for the vertices
 * @param faces the face halfedges
 */
inline void rotateHexZQ ( int *currentRotation, SVedge **faces )
{
	//rotate vertices
	int t = currentRotation[0];
	currentRotation[0] = currentRotation[3];
	currentRotation[3] = currentRotation[2];
	currentRotation[2] = currentRotation[1];
	currentRotation[1] = t;
	t = currentRotation[4];
	currentRotation[4] = currentRotation[7];
	currentRotation[7] = currentRotation[6];
	currentRotation[6] = currentRotation[5];
	currentRotation[5] = t;

	//rotate faces
	SVedge* f = faces[1];
	faces[1] = faces[4] ? faces[4]->next : NULL;
	faces[4] = faces[3];
	faces[3] = faces[2];
	faces[2] = f ? f->next->next->next : NULL;

	if(faces[0]) faces[0] = faces[0]->next;
	if(faces[5]) faces[5] = faces[5]->next->next->next;
}

/**
 * rotate the hexahedron a quarter (Q) of a turn about the Z axis in the inverse trigo direction
 * @param currentRotation the current rotation for the vertices
 * @param faces the face halfedges
 */
inline void rotateHexZIQ ( int *currentRotation, SVedge **faces )
{
	//rotate vertices
	int t = currentRotation[0];
	currentRotation[0] = currentRotation[1];
	currentRotation[1] = currentRotation[2];
	currentRotation[2] = currentRotation[3];
	currentRotation[3] = t;
	t = currentRotation[4];
	currentRotation[4] = currentRotation[5];
	currentRotation[5] = currentRotation[6];
	currentRotation[6] = currentRotation[7];
	currentRotation[7] = t;

	//rotate faces
	SVedge* f = faces[1];
	faces[1] = faces[2] ? faces[2]->next : NULL;
	faces[2] = faces[3];
	faces[3] = faces[4];
	faces[4] = f ? f->next->next->next : NULL;

	if(faces[0]) faces[0] = faces[0]->next->next->next;
	if(faces[5]) faces[5] = faces[5]->next;
}

/**
 * rotate the hexahedron a half (H) of a turn about the Z axis
 * @param currentRotation the current rotation for the vertices
 * @param faces the face halfedges
 */
inline void rotateHexZH ( int *currentRotation, SVedge **faces )
{
	//rotate vertices
	int t = currentRotation[0];
	currentRotation[0] = currentRotation[2];
	currentRotation[2] = t;
	t = currentRotation[1];
	currentRotation[1] = currentRotation[3];
	currentRotation[3] = t;
	t = currentRotation[4];
	currentRotation[4] = currentRotation[6];
	currentRotation[6] = t;
	t = currentRotation[5];
	currentRotation[5] = currentRotation[7];
	currentRotation[7] = t;

	//rotate faces
	SVedge* f = faces[1];
	faces[1] = faces[3] ? faces[3]->next : NULL;
	faces[3] = f ? f->next->next->next : NULL;
	f = faces[2];
	faces[2] = faces[4];
	faces[4] = f;

	if(faces[0]) faces[0] = faces[0]->next->next;
	if(faces[5]) faces[5] = faces[5]->next->next;
}


/**
 * rotate the hexahedron a quarter (Q) of a turn about the Y axis in the trigo direction
 * @param currentRotation the current rotation for the vertices
 * @param faces the face halfedges
 */
inline void rotateHexYQ ( int *currentRotation, SVedge **faces )
{
	//rotate vertices
	int t = currentRotation[0];
	currentRotation[0] = currentRotation[4];
	currentRotation[4] = currentRotation[7];
	currentRotation[7] = currentRotation[3];
	currentRotation[3] = t;
	t = currentRotation[1];
	currentRotation[1] = currentRotation[5];
	currentRotation[5] = currentRotation[6];
	currentRotation[6] = currentRotation[2];
	currentRotation[2] = t;

	//rotate faces
	SVedge* f = faces[0];
	faces[0] = faces[2] ? faces[2]->next->next->next : NULL;
	faces[2] = faces[5];
	faces[5] = faces[4] ? faces[4]->next->next : NULL;
	faces[4] = f ? f->next->next->next : NULL;

	if(faces[1]) faces[1] = faces[1]->next;
	if(faces[3]) faces[3] = faces[3]->next->next->next;
}

/**
 * rotate the hexahedron a quarter (Q) of a turn about the Y axis in the inverse trigo direction
 * @param currentRotation the current rotation for the vertices
 * @param faces the face halfedges
 */
inline void rotateHexYIQ ( int *currentRotation, SVedge **faces )
{
	//FIXME: performance: do this in one pass just like rotateHexZQ
	rotateHexYQ(currentRotation, faces);
	rotateHexYQ(currentRotation, faces);
	rotateHexYQ(currentRotation, faces);
}

/**
 * rotate the hexahedron a half (H) of a turn about the Y axis
 * @param currentRotation the current rotation for the vertices
 * @param faces the face halfedges
 */
inline void rotateHexYH ( int *currentRotation, SVedge **faces )
{
	//FIXME: performance: do this in one pass just like rotateHexZQ
	rotateHexYQ(currentRotation, faces);
	rotateHexYQ(currentRotation, faces);
}


/**
 * rotate the hexahedron a quarter (Q) of a turn about the X axis in the trigo direction
 * @param currentRotation the current rotation for the vertices
 * @param faces the face halfedges
 */
inline void rotateHexXQ ( int *currentRotation, SVedge **faces )
{
	//rotate vertices
	int t = currentRotation[0];
	currentRotation[0] = currentRotation[1];
	currentRotation[1] = currentRotation[5];
	currentRotation[5] = currentRotation[4];
	currentRotation[4] = t;
	t = currentRotation[2];
	currentRotation[2] = currentRotation[6];
	currentRotation[6] = currentRotation[7];
	currentRotation[7] = currentRotation[3];
	currentRotation[3] = t;

	//rotate faces
	SVedge* f = faces[0];
	faces[0] = faces[3];
	faces[3] = faces[5] ? faces[5]->next : NULL;
	faces[5] = faces[1];
	faces[1] = f ? f->next->next->next : NULL;

	if(faces[2]) faces[2] = faces[2]->next;
	if(faces[4]) faces[4] = faces[4]->next->next->next;
}

/**
 * rotate the hexahedron a quarter (Q) of a turn about the X axis in the inverse trigo direction
 * @param currentRotation the current rotation for the vertices
 * @param faces the face halfedges
 */
inline void rotateHexXIQ ( int *currentRotation, SVedge **faces )
{
	//FIXME: performance: do this in one pass just like rotateHexZQ
	rotateHexXQ(currentRotation, faces);
	rotateHexXQ(currentRotation, faces);
	rotateHexXQ(currentRotation, faces);
}

/**
 * rotate the hexahedron a half (H) of a turn about the X axis
 * @param currentRotation the current rotation for the vertices
 * @param faces the face halfedges
 */
inline void rotateHexXH ( int *currentRotation, SVedge **faces )
{
	//FIXME: performance: do this in one pass just like rotateHexZQ
	rotateHexXQ(currentRotation, faces);
	rotateHexXQ(currentRotation, faces);
}

/**
 * rotate the hexahedron to exchange F0 and F1
 * @param currentRotation the current rotation for the vertices
 * @param faces the face halfedges
 */
inline void rotateHexExchangeF0F1 ( int *currentRotation, SVedge **faces )
{
	//FIXME: performance: do this in one pass just like rotateHexZQ
	rotateHexZH(currentRotation, faces);
	rotateHexXQ(currentRotation, faces);
}

/**
 * rotate the hexahedron about the 0->6 axis a third of a turn in the trigo direction
 * @param currentRotation the current rotation for the vertices
 * @param faces the face halfedges
 */
inline void rotateHex06T ( int *currentRotation, SVedge **faces )
{
	//rotate vertices
	int t = currentRotation[1];
	currentRotation[1] = currentRotation[4];
	currentRotation[4] = currentRotation[3];
	currentRotation[3] = t;
	t = currentRotation[2];
	currentRotation[2] = currentRotation[5];
	currentRotation[5] = currentRotation[7];
	currentRotation[7] = t;

	//rotate faces
	SVedge* f = faces[0];
	faces[0] = faces[2];
	faces[2] = faces[1];
	faces[1] = f;
	f = faces[3];
	faces[3] = faces[5];
	faces[5] = faces[4] ? faces[4]->next : NULL;
	faces[4] = f ? f->next->next->next : NULL;
}


/**
 * rotate the hexahedron about the 0->6 axis a third of a turn in the trigo direction
 * @param currentRotation the current rotation for the vertices
 * @param faces the face halfedges
 */
inline void rotateHex06IT ( int *currentRotation, SVedge **faces )
{
	//rotate vertices
	int t = currentRotation[1];
	currentRotation[1] = currentRotation[3];
	currentRotation[3] = currentRotation[4];
	currentRotation[4] = t;
	t = currentRotation[2];
	currentRotation[2] = currentRotation[7];
	currentRotation[7] = currentRotation[5];
	currentRotation[5] = t;

	//rotate faces
	SVedge* f = faces[0];
	faces[0] = faces[1];
	faces[1] = faces[2];
	faces[2] = f;
	f = faces[3];
	faces[3] = faces[4] ? faces[4]->next : NULL;
	faces[4] = faces[5] ? faces[5]->next->next->next : NULL;
	faces[5] = f;
}



#if 0
inline void rotateAroundF0 ( int *currentRotation, int nSteps, SVedge **faceRotation=NULL )
	static const int f0Rotation[] = {3,0,1,2,7,4,5,6};
	static const int f0FaceRotation[] = {0,4,1,2,3,5};
	bool writeToTemp = true;
	int tCurrentRotation[8];
	SVedge* tFaceRotation[6];

	for ( int steps=0; steps<nSteps%4; ++steps )
	{
		if ( writeToTemp )
		{
			for ( int i=0; i<8; ++i )
				tCurrentRotation[i] = currentRotation[f0Rotation[i]];
			if ( faceRotation )
				for ( int i=0; i<6; ++i )
					tFaceRotation[i] = faceRotation[f0FaceRotation[i]];
			writeToTemp=false;
		}
		else
		{
			for ( int i=0; i<8; ++i )
				currentRotation[i] = tCurrentRotation[f0Rotation[i]];
			if ( faceRotation )
				for ( int i=0; i<6; ++i )
					faceRotation[i] = tFaceRotation[f0FaceRotation[i]];
			writeToTemp=true;
		}
	}
	if ( !writeToTemp )
	{
		for ( int i=0; i<8; ++i )
			currentRotation[i] = tCurrentRotation[i];
		if ( faceRotation )
			for ( int i=0; i<6; ++i )
				faceRotation[i] = tFaceRotation[i];
	}
}

inline void rotateAroundF1 ( int *currentRotation, int nSteps, SVedge **faceRotation=NULL )
{
	static int f1Rotation[] = {4,5,1,0,7,6,2,3};
	static const int f1FaceRotation[] = {2,1,5,3,0,4};
	bool writeToTemp = true;
	int tCurrentRotation[8];
	SVedge* tFaceRotation[6];

	for ( int steps=0; steps<nSteps%4; ++steps )
	{
		if ( writeToTemp )
		{
			for ( int i=0; i<8; ++i )
				tCurrentRotation[i] = currentRotation[f1Rotation[i]];
			if ( faceRotation )
				for ( int i=0; i<6; ++i )
					tFaceRotation[i] = faceRotation[f1FaceRotation[i]];
			writeToTemp=false;
		}
		else
		{
			for ( int i=0; i<8; ++i )
				currentRotation[i] = tCurrentRotation[f1Rotation[i]];
			if ( faceRotation )
				for ( int i=0; i<6; ++i )
					faceRotation[i] = tFaceRotation[f1FaceRotation[i]];
			writeToTemp=true;
		}
	}
	if ( !writeToTemp )
	{
		for ( int i=0; i<8; ++i )
			currentRotation[i] = tCurrentRotation[i];
		if ( faceRotation )
			for ( int i=0; i<6; ++i )
				faceRotation[i] = tFaceRotation[i];
	}
}

inline void rotateAroundF2 ( int *currentRotation, int nSteps, SVedge **faceRotation=NULL )
{
	static int f2Rotation[] = {1,5,6,2,0,4,7,3};
	static const int f2FaceRotation[] = {3,0,2,5,4,1};
	bool writeToTemp = true;
	int tCurrentRotation[8];
	SVedge* tFaceRotation[6];

	for ( int steps=0; steps<nSteps%4; ++steps )
	{
		if ( writeToTemp )
		{
			for ( int i=0; i<8; ++i )
				tCurrentRotation[i] = currentRotation[f2Rotation[i]];
			if ( faceRotation )
				for ( int i=0; i<6; ++i )
					tFaceRotation[i] = faceRotation[f2FaceRotation[i]];
			writeToTemp=false;
		}
		else
		{
			for ( int i=0; i<8; ++i )
				currentRotation[i] = tCurrentRotation[f2Rotation[i]];
			if ( faceRotation )
				for ( int i=0; i<6; ++i )
					faceRotation[i] = tFaceRotation[f2FaceRotation[i]];
			writeToTemp=true;
		}
	}
	if ( !writeToTemp )
	{
		for ( int i=0; i<8; ++i )
			currentRotation[i] = tCurrentRotation[i];
		if ( faceRotation )
			for ( int i=0; i<6; ++i )
				faceRotation[i] = tFaceRotation[i];
	}
}
#endif

#endif
