/*
===============================================================================

  FILE:  svcommon_svc.cpp
  
  CONTENTS:

    code that is common to svreader_svc.cpp and svwriter_svc.cpp

  PROGRAMMERS:

    martin isenburg@cs.unc.edu

  COPYRIGHT:

    copyright (C) 2010  martin isenburg@cs.unc.edu

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    25 January 2010 -- created behind trader joe's   

===============================================================================
*/

// shared defines

#define SVC_ADD_NORMAL_COMPONENT
#undef SVC_ADD_NORMAL_COMPONENT

#define SVC_USE_LOCAL_COORDINATE_SYSTEM
#undef SVC_USE_LOCAL_COORDINATE_SYSTEM

#define SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION
#undef SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION

#define SVC_DO_FLIP_PREDICTION
#undef SVC_DO_FLIP_PREDICTION

#define SVC_STORE_QUANTIZED_COORDINATES
#undef SVC_STORE_QUANTIZED_COORDINATES

#define SVC_DO_NOT_FINALIZE
#undef SVC_DO_NOT_FINALIZE

#define SVC_GUESS_FINALIZATION
#undef SVC_GUESS_FINALIZATION

#define SVC_START 0
#define SVC_ADD 1
#define SVC_JOIN 2
#define SVC_FLIP 3
#define SVC_FILL 4
#define SVC_END 5

#define SVC_MAX_DEGREE_ONE 10
#define SVC_MAX_USE_COUNT 18
#define SVC_INITIAL_VERTEX_LIST_SIZE 10
#define SVC_NORMAL_SCALE 0.8f

// shared includes

#include "rangefloatcompressor.h"
#include "rangeintegercompressor.h"
#include "positionquantizer_new.h"
#include "dynamicvector.h"
#include "littlecache.h"
#include "vec3fv.h"
#include "vec3iv.h"
#include "vecnfv.h"

// shared variables and functions

static DynamicVector* dv;
static LittleCache* lc;
static LittleCache* lc_f;

static RangeFloatCompressor* v_fc[3];
static PositionQuantizerNew* v_pq;
static RangeIntegerCompressor* global_ic[3];

#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
static float maxDist = 0.0f;
static int maxAbso = 0;
static RangeIntegerCompressor* local_ic[3];
#endif

#ifdef SVC_STORE_QUANTIZED_COORDINATES
static FILE* quantized_coordinates_file;
#endif

#ifdef SVC_ADD_NORMAL_COMPONENT
#endif // SVC_ADD_NORMAL_COMPONENT

// what was the last operation
static int last_op;

// context about the last finalization
static int last_fin;

// is there more to encode
static RangeModel* rmDone;

// codes next operation
static RangeModel** rmOp;
static RangeModel*** rmOpX;

// codes non-manifoldness of start operations
static RangeModel* rmS_Old;

// codes cache hits for start operations
static RangeModel** rmS_Cache;

// codes cache hits for add/join operations
static RangeModel** rmAJ_FaceCache;
static RangeModel*** rmAJ_FaceCacheX;
static RangeModel** rmAJ_EdgeCache;
static RangeModel** rmAJ_VertexCache;

static RangeModel** rmAJ_LocalJoin;

// codes cache hits fill/end operations
static RangeModel** rmFE_FaceCache;
static RangeModel*** rmFE_FaceCacheX;
static RangeModel** rmFE_EdgeCache;
static RangeModel** rmFE_VertexCache;

static RangeModel** rm_VertexCandi;
static RangeModel* rm_FaceCandi;

// codes vertex finalization
static RangeModel*** rmFinalized;

#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
static I32 quantize(float v)
{
  return (I32)floor((v*maxAbso+0.5*maxDist)/maxDist);
}
static float dequantize(I32 v)
{
  return (v*maxDist)/maxAbso;
}
#endif

// efficient memory allocation for vertices

static void initModels(int compress)
{
  int i,j;
  unsigned int t[] = {1,1,1,1,1,1,1,1,1,1,1,1,1};

  // range table for whether there is more
  rmDone = new RangeModel(2,0,compress);

  // range table for which operation it is
  rmOp = (RangeModel**)malloc(sizeof(RangeModel*)*6);
  rmOp[0] = new RangeModel(4,0,compress); // after SVC_START
  rmOp[1] = new RangeModel(4,0,compress); // after SVC_ADD
  rmOp[2] = new RangeModel(4,0,compress); // after SVC_JOIN
  rmOp[3] = new RangeModel(4,0,compress); // after SVC_FLIP
  rmOp[4] = new RangeModel(4,0,compress); // after SVC_FILL
  rmOp[5] = new RangeModel(4,0,compress); // after SVC_END

  // range table for which operation it is
  rmOpX = (RangeModel***)malloc(sizeof(RangeModel**)*20);
  for (i = 0; i < 20; i++)
  {
    rmOpX[i] = (RangeModel**)malloc(sizeof(RangeModel*)*6);
    rmOpX[i][0] = new RangeModel(4,0,compress); // after SVC_START
    rmOpX[i][1] = new RangeModel(4,0,compress); // after SVC_ADD
    rmOpX[i][2] = new RangeModel(4,0,compress); // after SVC_JOIN
    rmOpX[i][3] = new RangeModel(4,0,compress); // after SVC_FLIP
    rmOpX[i][4] = new RangeModel(4,0,compress); // after SVC_FILL
    rmOpX[i][5] = new RangeModel(4,0,compress); // after SVC_END
  }

  // range tables for start (S) operation
  rmS_Old = new RangeModel(6,0,compress); // number of old (e.g. non-manifold) vertices in a start operation (5 = NON_FINALIZED_EOF)
  rmS_Cache = (RangeModel**)malloc(sizeof(RangeModel*)*4);
  rmS_Cache[0] = new RangeModel(5,0,compress); // for 0
  rmS_Cache[1] = new RangeModel(5,0,compress); // for 1
  rmS_Cache[2] = new RangeModel(5,0,compress); // for 2
  rmS_Cache[3] = new RangeModel(5,0,compress); // for 3

  // range tables for add (A) / join (J) operations
  rmAJ_FaceCache = (RangeModel**)malloc(sizeof(RangeModel*)*6);
  t[0] = 1; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1;
  rmAJ_FaceCache[0] = new RangeModel(5,t,compress); // after SVC_START
  t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1;
  rmAJ_FaceCache[1] = new RangeModel(5,t,compress); // after SVC_ADD
  t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1;
  rmAJ_FaceCache[2] = new RangeModel(5,t,compress); // after SVC_JOIN
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 1; t[4] = 1;
  rmAJ_FaceCache[3] = new RangeModel(5,t,compress); // after SVC_FLIP
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 1; t[4] = 1;
  rmAJ_FaceCache[4] = new RangeModel(5,t,compress); // after SVC_FILL
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 1;
  rmAJ_FaceCache[5] = new RangeModel(5,t,compress); // after SVC_END

  // range tables for add (A) / join (J) operations
  rmAJ_FaceCacheX = (RangeModel***)malloc(sizeof(RangeModel**)*20);
  for (i = 0; i < 20; i++)
  {
    rmAJ_FaceCacheX[i] = (RangeModel**)malloc(sizeof(RangeModel*)*6);
    t[0] = 1; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1;
    rmAJ_FaceCacheX[i][0] = new RangeModel(5,t,compress); // after SVC_START
    t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1;
    rmAJ_FaceCacheX[i][1] = new RangeModel(5,t,compress); // after SVC_ADD
    t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1;
    rmAJ_FaceCacheX[i][2] = new RangeModel(5,t,compress); // after SVC_JOIN
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 1; t[4] = 1;
    rmAJ_FaceCacheX[i][3] = new RangeModel(5,t,compress); // after SVC_FLIP
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 1; t[4] = 1;
    rmAJ_FaceCacheX[i][4] = new RangeModel(5,t,compress); // after SVC_FILL
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 1;
    rmAJ_FaceCacheX[i][5] = new RangeModel(5,t,compress); // after SVC_END
  }

  rmAJ_EdgeCache = (RangeModel**)malloc(sizeof(RangeModel*)*6);
  rmAJ_EdgeCache[0] = new RangeModel(14,0,compress); // after SVC_START
  rmAJ_EdgeCache[1] = new RangeModel(14,0,compress); // after SVC_ADD
  rmAJ_EdgeCache[2] = new RangeModel(14,0,compress); // after SVC_JOIN
  rmAJ_EdgeCache[3] = new RangeModel(14,0,compress); // after SVC_FLIP
  rmAJ_EdgeCache[4] = new RangeModel(14,0,compress); // after SVC_FILL
  rmAJ_EdgeCache[5] = new RangeModel(14,0,compress); // after SVC_END

  rmAJ_VertexCache = (RangeModel**)malloc(sizeof(RangeModel*)*6);
  rmAJ_VertexCache[0] = new RangeModel(5,0,compress); // after SVC_START
  rmAJ_VertexCache[1] = new RangeModel(5,0,compress); // after SVC_ADD
  rmAJ_VertexCache[2] = new RangeModel(5,0,compress); // after SVC_JOIN
  rmAJ_VertexCache[3] = new RangeModel(5,0,compress); // after SVC_FLIP
  rmAJ_VertexCache[4] = new RangeModel(5,0,compress); // after SVC_FILL
  rmAJ_VertexCache[5] = new RangeModel(5,0,compress); // after SVC_END

  rmAJ_LocalJoin = (RangeModel**)malloc(sizeof(RangeModel*)*6);
  rmAJ_LocalJoin[0] = new RangeModel(7,0,compress); // after SVC_START
  rmAJ_LocalJoin[1] = new RangeModel(7,0,compress); // after SVC_ADD
  rmAJ_LocalJoin[2] = new RangeModel(7,0,compress); // after SVC_JOIN
  rmAJ_LocalJoin[3] = new RangeModel(7,0,compress); // after SVC_FLIP
  rmAJ_LocalJoin[4] = new RangeModel(7,0,compress); // after SVC_FILL
  rmAJ_LocalJoin[5] = new RangeModel(7,0,compress); // after SVC_END

  // range tables for fill (F) / end (E) operations

  rmFE_FaceCache = (RangeModel**)malloc(sizeof(RangeModel*)*6);
  t[0] = 1; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
  rmFE_FaceCache[0] = new RangeModel(13,t,compress); // after SVC_START
  t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 0; t[5] = 1; t[6] = 1; t[7] = 1; t[8] = 0; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
  rmFE_FaceCache[1] = new RangeModel(13,t,compress); // after SVC_ADD
  t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 0; t[5] = 1; t[6] = 1; t[7] = 1; t[8] = 0; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
  rmFE_FaceCache[2] = new RangeModel(13,t,compress); // after SVC_JOIN
  t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 1; t[4] = 0; t[5] = 0; t[6] = 1; t[7] = 1; t[8] = 0; t[9] = 0; t[10] = 1; t[11] = 1; t[12] = 1;
  rmFE_FaceCache[3] = new RangeModel(13,t,compress); // after SVC_FLIP
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 1; t[4] = 0; t[5] = 0; t[6] = 0; t[7] = 1; t[8] = 0; t[9] = 0; t[10] = 0; t[11] = 1; t[12] = 1;
  rmFE_FaceCache[4] = new RangeModel(13,t,compress); // after SVC_FILL
  t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 0; t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 0; t[11] = 0; t[12] = 1;
  rmFE_FaceCache[5] = new RangeModel(13,t,compress); // after SVC_END

  rmFE_FaceCacheX = (RangeModel***)malloc(sizeof(RangeModel**)*20);
  for (i = 0; i < 20; i++)
  {
    rmFE_FaceCacheX[i] = (RangeModel**)malloc(sizeof(RangeModel*)*6);
    t[0] = 1; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 1; t[5] = 1; t[6] = 1; t[7] = 1; t[8] = 1; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
    rmFE_FaceCacheX[i][0] = new RangeModel(13,t,compress); // after SVC_START
    t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 0; t[5] = 1; t[6] = 1; t[7] = 1; t[8] = 0; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
    rmFE_FaceCacheX[i][1] = new RangeModel(13,t,compress); // after SVC_ADD
    t[0] = 0; t[1] = 1; t[2] = 1; t[3] = 1; t[4] = 0; t[5] = 1; t[6] = 1; t[7] = 1; t[8] = 0; t[9] = 1; t[10] = 1; t[11] = 1; t[12] = 1;
    rmFE_FaceCacheX[i][2] = new RangeModel(13,t,compress); // after SVC_JOIN
    t[0] = 0; t[1] = 0; t[2] = 1; t[3] = 1; t[4] = 0; t[5] = 0; t[6] = 1; t[7] = 1; t[8] = 0; t[9] = 0; t[10] = 1; t[11] = 1; t[12] = 1;
    rmFE_FaceCacheX[i][3] = new RangeModel(13,t,compress); // after SVC_FLIP
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 1; t[4] = 0; t[5] = 0; t[6] = 0; t[7] = 1; t[8] = 0; t[9] = 0; t[10] = 0; t[11] = 1; t[12] = 1;
    rmFE_FaceCacheX[i][4] = new RangeModel(13,t,compress); // after SVC_FILL
    t[0] = 0; t[1] = 0; t[2] = 0; t[3] = 0; t[4] = 0; t[5] = 0; t[6] = 0; t[7] = 0; t[8] = 0; t[9] = 0; t[10] = 0; t[11] = 0; t[12] = 1;
    rmFE_FaceCacheX[i][5] = new RangeModel(13,t,compress); // after SVC_END
  }

  rmFE_EdgeCache = (RangeModel**)malloc(sizeof(RangeModel*)*6);
  rmFE_EdgeCache[0] = new RangeModel(28,0,compress); // after SVC_START
  rmFE_EdgeCache[1] = new RangeModel(28,0,compress); // after SVC_ADD
  rmFE_EdgeCache[2] = new RangeModel(28,0,compress); // after SVC_JOIN
  rmFE_EdgeCache[3] = new RangeModel(28,0,compress); // after SVC_FLIP
  rmFE_EdgeCache[4] = new RangeModel(28,0,compress); // after SVC_FILL
  rmFE_EdgeCache[5] = new RangeModel(28,0,compress); // after SVC_END

  rmFE_VertexCache = (RangeModel**)malloc(sizeof(RangeModel*)*6);
  rmFE_VertexCache[0] = new RangeModel(9,0,compress); // after SVC_START
  rmFE_VertexCache[1] = new RangeModel(9,0,compress); // after SVC_ADD
  rmFE_VertexCache[2] = new RangeModel(9,0,compress); // after SVC_JOIN
  rmFE_VertexCache[3] = new RangeModel(9,0,compress); // after SVC_FLIP
  rmFE_VertexCache[4] = new RangeModel(9,0,compress); // after SVC_FILL
  rmFE_VertexCache[5] = new RangeModel(9,0,compress); // after SVC_END

  rm_VertexCandi = (RangeModel**)malloc(sizeof(RangeModel*)*14);
  for (i = 2; i < 14; i++)
  {
    rm_VertexCandi[i] = new RangeModel(i,0,compress);
  }

  rm_FaceCandi = new RangeModel(2,0,compress);

  rmFinalized = (RangeModel***)malloc(sizeof(RangeModel**)*SVC_MAX_DEGREE_ONE);
  for (i = 0; i < SVC_MAX_DEGREE_ONE; i++)
  {
    rmFinalized[i] = (RangeModel**)malloc(sizeof(RangeModel*)*SVC_MAX_USE_COUNT);
    for (j = 1; j < SVC_MAX_USE_COUNT; j++)
    {
      rmFinalized[i][j] = new RangeModel(2,0,compress);
    }
  }
}

static void finishModels()
{
  int i,j;

  // range table for whether there is more
  delete rmDone;

  // range table for which operation it is
  delete rmOp[0];
  delete rmOp[1];
  delete rmOp[2];
  delete rmOp[3];
  delete rmOp[4];
  delete rmOp[5];
  free(rmOp);

  // range table for which operation it is
  for (i = 0; i < 20; i++)
  {
    delete rmOpX[i][0];
    delete rmOpX[i][1];
    delete rmOpX[i][2];
    delete rmOpX[i][3];
    delete rmOpX[i][4];
    delete rmOpX[i][5];
    free(rmOpX[i]);
  }
  free (rmOpX);

  // range tables for start (S) operation
  delete rmS_Old;

  // range tables for start (S) operations
  delete rmS_Cache[0];
  delete rmS_Cache[1];
  delete rmS_Cache[2];
  delete rmS_Cache[3];
  free(rmS_Cache);

  // range tables for add (A) / join (J) operations
  delete rmAJ_FaceCache[0];
  delete rmAJ_FaceCache[1];
  delete rmAJ_FaceCache[2];
  delete rmAJ_FaceCache[3];
  delete rmAJ_FaceCache[4];
  delete rmAJ_FaceCache[5];
  free(rmAJ_FaceCache);
  for (i = 0; i < 20; i++)
  {
    delete rmAJ_FaceCacheX[i][0];
    delete rmAJ_FaceCacheX[i][1];
    delete rmAJ_FaceCacheX[i][2];
    delete rmAJ_FaceCacheX[i][3];
    delete rmAJ_FaceCacheX[i][4];
    delete rmAJ_FaceCacheX[i][5];
    free(rmAJ_FaceCacheX[i]);
  }
  free(rmAJ_FaceCacheX);
  delete rmAJ_EdgeCache[0];
  delete rmAJ_EdgeCache[1];
  delete rmAJ_EdgeCache[2];
  delete rmAJ_EdgeCache[3];
  delete rmAJ_EdgeCache[4];
  delete rmAJ_EdgeCache[5];
  free(rmAJ_EdgeCache);
  delete rmAJ_VertexCache[0];
  delete rmAJ_VertexCache[1];
  delete rmAJ_VertexCache[2];
  delete rmAJ_VertexCache[3];
  delete rmAJ_VertexCache[4];
  delete rmAJ_VertexCache[5];
  free(rmAJ_VertexCache);
  delete rmAJ_LocalJoin[0];
  delete rmAJ_LocalJoin[1];
  delete rmAJ_LocalJoin[2];
  delete rmAJ_LocalJoin[3];
  delete rmAJ_LocalJoin[4];
  delete rmAJ_LocalJoin[5];
  free(rmAJ_LocalJoin);

  // range tables for fill (F) / end (E) operations
  delete rmFE_FaceCache[0];
  delete rmFE_FaceCache[1];
  delete rmFE_FaceCache[2];
  delete rmFE_FaceCache[3];
  delete rmFE_FaceCache[4];
  delete rmFE_FaceCache[5];
  free(rmFE_FaceCache);
  for (i = 0; i < 20; i++)
  {
    delete rmFE_FaceCacheX[i][0];
    delete rmFE_FaceCacheX[i][1];
    delete rmFE_FaceCacheX[i][2];
    delete rmFE_FaceCacheX[i][3];
    delete rmFE_FaceCacheX[i][4];
    delete rmFE_FaceCacheX[i][5];
    free(rmFE_FaceCacheX[i]);
  }
  free(rmFE_FaceCacheX);
  delete rmFE_EdgeCache[0];
  delete rmFE_EdgeCache[1];
  delete rmFE_EdgeCache[2];
  delete rmFE_EdgeCache[3];
  delete rmFE_EdgeCache[4];
  delete rmFE_EdgeCache[5];
  free(rmFE_EdgeCache);
  delete rmFE_VertexCache[0];
  delete rmFE_VertexCache[1];
  delete rmFE_VertexCache[2];
  delete rmFE_VertexCache[3];
  delete rmFE_VertexCache[4];
  delete rmFE_VertexCache[5];
  free(rmFE_VertexCache);

  for (i = 2; i < 14; i++)
  {
    delete rm_VertexCandi[i];
  }
  free(rm_VertexCandi);

  delete rm_FaceCandi;

  for (i = 0; i < SVC_MAX_DEGREE_ONE; i++)
  {
    for (j = 1; j < SVC_MAX_USE_COUNT; j++)
    {
      delete rmFinalized[i][j];
    }
    free(rmFinalized[i]);
  }
  free(rmFinalized);
}

// efficient memory allocation

struct SVedge;

typedef struct SVvertex
{
  SVvertex* buffer_next;    // used for efficient memory management and by dynamicvector data structure
  float* prop_f;
  I32 index;
  I32 use_count;
  I32 in_cache;
  I32 finalize;
  int list_size;
  int list_alloc;
  SVedge** list;
} SVvertex;

static int vertex_pnum_f = 0;
static SVvertex* vertex_buffer_next;
static int vertex_buffer_increment;
static int vertex_buffer_number_chunks;
static SVvertex* vertex_buffer_chunks[16];
#ifdef PRINT_CONTROL_OUTPUT
static int vertex_buffer_size;
static int vertex_buffer_maxsize;
static int vertex_buffer_alloc;
#endif

static void initVertexBuffer(int alloc, int pnum_f)
{
  vertex_pnum_f = pnum_f;
  vertex_buffer_next = 0;
  vertex_buffer_increment = alloc;
  vertex_buffer_number_chunks = 0;
#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size = 0;
  vertex_buffer_maxsize = 0;
  vertex_buffer_alloc = 0;
#endif
}

static SVvertex* allocVertex()
{
  if ( vertex_buffer_next == 0 )
  {
    int increment = vertex_buffer_increment;
    vertex_buffer_increment *= 2;
    vertex_buffer_next = ( SVvertex* ) malloc ( sizeof ( SVvertex ) * increment );
    if ( vertex_buffer_next == 0 )
    {
      fprintf ( stderr,"ERROR (SVreader_svh): malloc for vertex buffer %d failed\n", increment );
      return 0;
    }
    else
    {
      vertex_buffer_chunks[vertex_buffer_number_chunks] = vertex_buffer_next;
      vertex_buffer_number_chunks++;      
    }
#ifdef PRINT_CONTROL_OUTPUT
    vertex_buffer_alloc += increment;
#endif

    if ( vertex_pnum_f )
    {
      vertex_buffer_next[0].prop_f = ( float* ) malloc ( sizeof ( float ) * increment * vertex_pnum_f );
      if ( vertex_buffer_next[0].prop_f == 0 )
      {
        fprintf ( stderr,"ERROR (SVreader_svh): malloc for vertex_buffer_next->prop_f %d failed\n", increment );
        return 0;
      }
    }
    for ( int i = 0; i < increment; i++ )
    {
      vertex_buffer_next[i].buffer_next = & ( vertex_buffer_next[i+1] );
      vertex_buffer_next[i].list = 0;
      if ( vertex_pnum_f )
      {
        vertex_buffer_next[i].prop_f = & ( vertex_buffer_next[0].prop_f[i*vertex_pnum_f] );
      }
    }
    vertex_buffer_next[increment-1].buffer_next = 0;
  }
  // get pointer to next available vertex
  SVvertex* vertex = vertex_buffer_next;
  vertex_buffer_next = vertex->buffer_next;

  if ( vertex->list == 0 )
  {
    vertex->list = ( SVedge** ) malloc ( sizeof ( SVedge* ) * SVC_INITIAL_VERTEX_LIST_SIZE );
    vertex->list_alloc = SVC_INITIAL_VERTEX_LIST_SIZE;
  }
  vertex->list_size = 0;
  vertex->use_count = 0;

#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size++;
  if ( vertex_buffer_size > vertex_buffer_maxsize ) vertex_buffer_maxsize = vertex_buffer_size;
#endif

  return vertex;
}

static void deallocVertex ( SVvertex* vertex )
{
  vertex->buffer_next = vertex_buffer_next;
  vertex_buffer_next = vertex;
#ifdef PRINT_CONTROL_OUTPUT
  vertex_buffer_size--;
#endif
}

static void destroyVertexBuffer()
{
  for (int i = vertex_buffer_number_chunks-1; i >= 0; i--)
  {
    vertex_buffer_increment/=2;
    for (int j = 0; j < vertex_buffer_increment; j++)
    {
      if (vertex_buffer_chunks[i][j].list) free(vertex_buffer_chunks[i][j].list);
    }
    if ( vertex_pnum_f )
    {
      free(vertex_buffer_chunks[i][0].prop_f);
    }
    free(vertex_buffer_chunks[i]);
  }
  vertex_buffer_number_chunks = 0;
  vertex_buffer_next = 0;
}

typedef struct SVedge     // this is an oriented half-edge
{
  SVedge* buffer_next;    // used for efficient memory management
  SVvertex* target;       // target vertex of oriented half-edge
  SVedge* next;           // next half-edge in ccw-oriented chain of half-edges around the face
#if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION) || defined(SVC_DO_FLIP_PREDICTION)
  float across[3];
#endif
} SVedge;

static SVedge* edge_buffer_next;
static int edge_buffer_increment;
static int edge_buffer_number_chunks;
static SVedge* edge_buffer_chunks[16];
static int edge_buffer_size;
#ifdef PRINT_CONTROL_OUTPUT
static int edge_buffer_maxsize;
static int edge_buffer_alloc;
#endif

static void initEdgeBuffer(int increment)
{
  edge_buffer_next = 0;
  edge_buffer_increment = increment;
  edge_buffer_number_chunks = 0;
  edge_buffer_size = 0;
#ifdef PRINT_CONTROL_OUTPUT
  edge_buffer_maxsize = 0;
  edge_buffer_alloc = 0;
#endif
}

static SVedge* allocEdge()
{
  if ( edge_buffer_next == 0 )
  {
    int increment = edge_buffer_increment;
    edge_buffer_increment *= 2;
    edge_buffer_next = ( SVedge* ) malloc ( sizeof ( SVedge ) * increment );
    if ( edge_buffer_next == 0 )
    {
      fprintf ( stderr,"malloc for edge buffer failed\n" );
      return 0;
    }
    else
    {
      edge_buffer_chunks[edge_buffer_number_chunks] = edge_buffer_next;
      edge_buffer_number_chunks++;      
    }
#ifdef PRINT_CONTROL_OUTPUT
    edge_buffer_alloc += increment;
#endif
    for ( int i = 0; i < increment; i++ )
    {
      edge_buffer_next[i].buffer_next = & ( edge_buffer_next[i+1] );
    }
    edge_buffer_next[increment-1].buffer_next = 0;
  }
  // get index of next available edge
  SVedge* edge = edge_buffer_next;
  edge_buffer_next = edge->buffer_next;

  edge->buffer_next = 0;
  edge->target = 0;
  edge->next = 0;

  edge_buffer_size++;

#ifdef PRINT_CONTROL_OUTPUT
  if ( edge_buffer_size > edge_buffer_maxsize ) edge_buffer_maxsize = edge_buffer_size;
#endif

  return edge;
}

static void deallocEdge ( SVedge* edge )
{
  edge->buffer_next = edge_buffer_next;
  edge_buffer_next = edge;
  edge_buffer_size--;
}

static void destroyEdgeBuffer()
{
  for (int i = edge_buffer_number_chunks-1; i >= 0; i--)
  {
    edge_buffer_increment/=2;
    free(edge_buffer_chunks[i]);
  }
  edge_buffer_number_chunks = 0;
  edge_buffer_next = 0;
}

static void addEdgeToVertex(SVedge* edge, SVvertex* vertex)
{
  if (vertex->list_size == vertex->list_alloc)
  {
    vertex->list_alloc += 2;
    vertex->list = (SVedge**)realloc(vertex->list, sizeof(SVedge*)*vertex->list_alloc);
  }
  vertex->list[vertex->list_size] = edge;
  vertex->list_size++;
}

static void removeEdgeFromVertex(SVedge* edge, SVvertex* vertex)
{
  int i;
  for (i = 0; i < vertex->list_size; i++)
  {
    if (vertex->list[i] == edge)
    {
      vertex->list_size--;
      vertex->list[i] = vertex->list[vertex->list_size];
      return;
    }
  }
  fprintf(stderr,"FATAL ERROR: did not find edge in vertex list\n");
  exit(0);
}
