/*
===============================================================================

  FILE:  SVwriter_svb.cpp
  
  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    martin isenburg@cs.unc.edu

  COPYRIGHT:

    copyright (C) 2004-07 martin isenburg@cs.unc.edu

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svwriter_svb.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>

#include "vec3fv.h"
#include "vec3iv.h"

#include "vecnfv.h"
#include "vecniv.h"

#define SV_VERSION 4 // this is the first SVB implementation

bool SVwriter_svb::open(FILE* file)
{
  if (file == 0)
  {
    fprintf(stderr, "ERROR: zero file pointer not supported by SVwriter_svb\n");
    exit(0);
  }

#ifdef _WIN32
  if (file == stdout)
  {
    if(_setmode( _fileno( stdout ), _O_BINARY ) == -1 )
    {
      fprintf(stderr, "ERROR: cannot set stdout to binary (untranslated) mode\n");
      exit(0);
    }
  }
#endif

  this->file = file;
  // write version
  fputc(SV_VERSION, file);

  v_count = 0;
  c_count = 0;

  element_number = 0;
  element_descriptor = 0;

  return true;
}

void SVwriter_svb::close(bool close_file, bool update)
{
  // close of SVwriter_svb
  write_buffer_remaining();
  if (nverts != v_count || ncells != c_count)
  {
    if (update && file != stdout) update_header();
  }
  if (close_file && file && file != stdout) fclose(file);
  file = 0;

  // close of SVwriter interface
  if (nverts != -1) if (nverts != v_count)  fprintf(stderr,"WARNING: set nverts %d but v_count %d\n",nverts,v_count);
  if (ncells != -1) if (ncells != c_count)  fprintf(stderr,"WARNING: set ncells %d but c_count %d\n",ncells,c_count);
  nverts = v_count;
  ncells = c_count;
  v_count = -1;
  c_count = -1;
}

void SVwriter_svb::set_endianness(bool big_endian)
{
#if (defined(i386) || defined(WIN32))   // if little endian machine
  endian_swap = big_endian;
#else                                   // else big endian machine
  endian_swap = !big_endian;
#endif
}

void SVwriter_svb::write_vertex(const float* v_pos_f)
{
  if (v_count + c_count == 0) write_header();

  if (endian_swap) VecCopyNfv_swap_endian((float*)&(element_buffer[element_number*4]), v_pos_f, v_pnum_f);
  else VecCopyNfv((float*)&(element_buffer[element_number*4]), v_pos_f, v_pnum_f);

  element_descriptor = 0x80000000 | (element_descriptor >> 1);
  element_number++;

  if (element_number == 32) write_buffer();

  v_count++;
}

void SVwriter_svb::write_tetrahedron(const I32* c_idx, const bool* c_final)
{
  if (v_count + c_count == 0) write_header();

  int c_idx_finalized[]  = { (c_final[0] ? c_idx[0]-v_count : c_idx[0]+1),(c_final[1] ? c_idx[1]-v_count : c_idx[1]+1), (c_final[2] ? c_idx[2]-v_count : c_idx[2]+1), (c_final[3] ? c_idx[3]-v_count : c_idx[3]+1) };
  if (endian_swap) VecCopyNiv_swap_endian((int*)&(element_buffer[element_number*4]), c_idx_finalized, 4);
  else VecCopyNiv((int*)&(element_buffer[element_number*4]), c_idx_finalized, 4);

  element_descriptor = (element_descriptor >> 1);
  element_number++;

  if (element_number == 32) write_buffer();

  c_count++;
}

void SVwriter_svb::write_tetrahedron(const I32* c_idx, const float*, const bool* c_final)
{
  write_tetrahedron(c_idx, c_final);
}


void SVwriter_svb::write_hexahedron(const I32* c_idx, const bool* c_final)
{
    fprintf(stderr, "WARNING: SVwriter_svb does not implement write_hexahedron()\n");
}

void SVwriter_svb::write_hexahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
    fprintf(stderr, "WARNING: SVwriter_svb does not implement write_hexahedron()\n");
}

void SVwriter_svb::write_finalized(I32 final_idx)
{
  fprintf(stderr, "ERROR: write_finalized(int final_idx) not supported by SVwriter_svb\n");
  exit(0);
}

#define SV_LITTLE_ENDIAN 0
#define SV_BIG_ENDIAN 1
#define SV_COMPRESSION 0

static int swap_endian_int(I32 input)
{
  int output;
  ((char*)&output)[0] = ((char*)&input)[3];
  ((char*)&output)[1] = ((char*)&input)[2];
  ((char*)&output)[2] = ((char*)&input)[1];
  ((char*)&output)[3] = ((char*)&input)[0];
  return output;
}

static unsigned int swap_endian_uint(U32 input)
{
  int output;
  ((char*)&output)[0] = ((char*)&input)[3];
  ((char*)&output)[1] = ((char*)&input)[2];
  ((char*)&output)[2] = ((char*)&input)[1];
  ((char*)&output)[3] = ((char*)&input)[0];
  return output;
}

void SVwriter_svb::write_header()
{
  I32 output;
  // endianness
#if (defined(i386) || defined(WIN32))   // if little endian machine
  if (endian_swap) fputc(SV_BIG_ENDIAN, file);
  else fputc(SV_LITTLE_ENDIAN, file);
#else                                    // else big endian machine
  if (endian_swap) fputc(SV_LITTLE_ENDIAN, file);
  else fputc(SV_BIG_ENDIAN, file);
#endif
  // compression
  fputc(SV_COMPRESSION, file);
  fputc(SV_COMPRESSION, file);
  // write comments
  if (endian_swap) output = swap_endian_int(ncomments);
  else output = ncomments;
  fwrite(&output, sizeof(int), 1, file);
  if (ncomments)
  {
    for (int i = 0; i < ncomments; i++)
    {
      if (endian_swap) output = swap_endian_int(strlen(comments[i]));
      else output = strlen(comments[i]);
      fwrite(&output, sizeof(I32), 1, file);
      fwrite(comments[i], sizeof(char), output, file);
    }
  }
  // record where nverts and nfaces are located
  update_seek_pos = (int)ftell(file);
  // write nverts
  if (endian_swap) output = swap_endian_int(nverts);
  else output = nverts;
  fwrite(&output, sizeof(I32), 1, file);
  // write ncells
  if (endian_swap) output = swap_endian_int(ncells);
  else output = ncells;
  fwrite(&output, sizeof(I32), 1, file);
  // write how many floating-point attributes (per vertex)
  if (endian_swap) output = swap_endian_int(v_pnum_f);
  else output = v_pnum_f;
  fwrite(&output, sizeof(I32), 1, file);
  // write how many integer attributes (per vertex)
  if (endian_swap) output = swap_endian_int(0);
  else output = 0;
  fwrite(&output, sizeof(I32), 1, file);
  // write bounding box of floating-point attributes (per vertex)
  if (v_pnum_f)
  {
    if (v_pmin_f && v_pmax_f)
    {
      fputc(1, file);
      if (endian_swap)
      {
        float temp[4];
        VecCopyNfv_swap_endian(temp, v_pmin_f, v_pnum_f);
        fwrite(temp, sizeof(float), v_pnum_f, file);
        VecCopyNfv_swap_endian(temp, v_pmax_f, v_pnum_f);
        fwrite(temp, sizeof(float), v_pnum_f, file);
      }
      else
      {
        fwrite(v_pmin_f, sizeof(float), v_pnum_f, file);
        fwrite(v_pmax_f, sizeof(float), v_pnum_f, file);
      }
    }
    else
    {
      fputc(0, file);
    }
  }
  // write bounding box of integer attributes (per vertex)
  if (0) // v_pnum_i
  {
    if (0)
    {
      fputc(1, file);
//      fwrite(v_pmin_i, sizeof(int), v_pnum_i, file);
//      fwrite(v_pmax_i, sizeof(int), v_pnum_i, file);
    }
    else
    {
      fputc(0, file);
    }
  }
}

void SVwriter_svb::update_header()
{
  if (file == stdout)
  {
    fprintf(stderr,"WARNING: cannot update header when writing to stdout\n");
    return;
  }
  if (fseek(file, update_seek_pos, SEEK_SET) != 0)
  {
    fprintf(stderr,"WARNING: cannot update header because fseek failed\n");
    return;
  }
  I32 output;
  // write nverts
  if (endian_swap) output = swap_endian_int(v_count);
  else output = v_count;
  fwrite(&output, sizeof(I32), 1, file);
  // write ncells
  if (endian_swap) output = swap_endian_int(c_count);
  else output = c_count;
  fwrite(&output, sizeof(I32), 1, file);
  if (fseek(file, 0L, SEEK_END) != 0)
  {
    fprintf(stderr,"WARNING: header was updated but final fseek failed\n");
  }
}

void SVwriter_svb::write_buffer()
{
  if (endian_swap) element_descriptor = swap_endian_uint(element_descriptor);
  fwrite(&element_descriptor, sizeof(U32), 1, file);
  element_descriptor = 0;
  fwrite(element_buffer, sizeof(I32), 32*4, file);
  element_number = 0;
}

void SVwriter_svb::write_buffer_remaining()
{
  element_descriptor = element_descriptor >> (32 - element_number);
  if (endian_swap) element_descriptor = swap_endian_uint(element_descriptor);
  fwrite(&element_descriptor, sizeof(U32), 1, file);
  element_descriptor = 0;
  fwrite(element_buffer, sizeof(I32), element_number*4, file);
  element_number = 0;
}

SVwriter_svb::SVwriter_svb()
{
  // init of SVwriter_svb
  file = 0;
  element_buffer = (I32*)malloc(sizeof(I32)*4*32);
  endian_swap = false;
}

SVwriter_svb::~SVwriter_svb()
{
  // clean-up for SVwriter_svb
  free(element_buffer);
}
